#include "server.h"
#include "send.h"
#include "list.h"
#include <ortp/ortp.h>
#include "ortp.h"

#define TARGET_ADDR(oc) inet_ntoa((oc)->addr)

extern struct gengetopt_args_info conf;

struct ortp_target {
	struct in_addr addr;
	int status;
	struct list_head node;
	int port;
	RtpSession *session;
};

static int numtargets;
static struct list_head targets;
static struct sender *self;

static void ortp_delete_target(struct ortp_target *ot, const char *msg)
{
	PARA_INFO_LOG("shutting down %s:%d (%s)\n", TARGET_ADDR(ot),
		ot->port, msg);
	numtargets--;
	list_del(&ot->node);
	free(ot);
	return;
}

static int ortp_send_buf(struct ortp_target *ot, const char *buf, int len, int ts)
{
	int ret = rtp_session_send_with_ts(ot->session, buf, len, ts);
	if (ret < 0)
		ortp_delete_target(ot, "send error");
	return ret;
}

static int ortp_send_ctrlmsg(struct ortp_target *ot, char *buf, int len, int ts)
{
	rtp_session_set_payload_type(ot->session, PAYLOAD_OTHER);
	if (ortp_send_buf(ot, buf, len, ts) < 0)
		return -1;

	rtp_session_set_payload_type(ot->session, PAYLOAD_AUDIO_CONTINUOUS);
	return 1;
}

static int ortp_send_notifier(struct ortp_target *ot, int code, int ts)
{
	char buf[2];

	PARA_INFO_LOG("sending notifier #%d\n", code);
	buf[0] = code;
	buf[1] = '\0';
	return ortp_send_ctrlmsg(ot, buf, 2, ts);
}

static int ortp_send_header(struct audio_format *af)
{
	int len;
	char *header, *tmpbuf;
	struct ortp_target *ot, *tmp;

	if (!af || !af->get_header_info)
		return 1;
	if (list_empty(&targets))
		return 1;
	header = af->get_header_info(&len);
	if (!header || len <= 0)
		return -1;
	tmpbuf = para_malloc(len + 1);
	*tmpbuf = ORTP_HEADER;
	memcpy(tmpbuf + 1, header, len);
	list_for_each_entry_safe(ot, tmp, &targets, node) {
		if (!ot->session)
			continue;
		int ts = rtp_session_get_current_send_ts(ot->session);
		ortp_send_ctrlmsg(ot, tmpbuf, len + 1, ts);
	}
	free(tmpbuf);
	return 1;
}

static int ortp_init_session(struct ortp_target *ot)
{
	RtpSession *s;

	PARA_NOTICE_LOG("sending to udp %s:%d\n", TARGET_ADDR(ot), ot->port);
	ot->session = rtp_session_new(RTP_SESSION_SENDONLY);
	s = ot->session;
	rtp_session_set_scheduling_mode(s, 1);
	rtp_session_set_blocking_mode(s, 1);
	rtp_session_set_remote_addr(s, TARGET_ADDR(ot), ot->port);
//	rtp_session_set_payload_type(s, PAYLOAD_AUDIO_CONTINUOUS);
	return 1;
}

static void ortp_end_session(struct ortp_target *ot)
{
	int ts;

	if (!ot->session)
		return;
	ts = rtp_session_get_current_send_ts(ot->session);
	if (ortp_send_notifier(ot, ORTP_EOF, ts) < 0)
		return;
	PARA_NOTICE_LOG("destroying session %s:%d\n",
		 TARGET_ADDR(ot), ot->port);
	rtp_session_destroy(ot->session);
	ot->session = NULL;
}

/* called by afs */
static void ortp_shutdown_targets(void)
{
	struct ortp_target *ot, *tmp;
	list_for_each_entry_safe(ot, tmp, &targets, node)
		ortp_end_session(ot);
}

static int ortp_send(struct audio_format *af, long unsigned current_chunk,
		long unsigned chunks_sent, const char *buf, size_t len)
{
	struct ortp_target *ot, *tmp;
	int i;

	if (self->status != SENDER_ON)
		return 1;
//	PARA_DEBUG_LOG("len: %d\n", len);
	i = -1;
	list_for_each_entry_safe(ot, tmp, &targets, node) {
		int bytes, ts;
		i++;
		if (!ot->session) {
			ortp_init_session(ot);
			if (!ot->session)
				continue;
		}
		ts = rtp_session_get_current_send_ts(ot->session);
		if (!current_chunk) {
			if (ortp_send_notifier(ot, ORTP_BOF, ts) < 0)
				continue;
		}
		if (!chunks_sent)
			ortp_send_header(af);
		bytes = ortp_send_buf(ot, buf, len, ts);
	}
	return 1;
}

static int ortp_com_on(struct sender_command_data *scd)
{
	self->status = SENDER_ON;
	return 1;
}

static int ortp_com_off(struct sender_command_data *scd)
{
	ortp_shutdown_targets();
	self->status = SENDER_OFF;
	return 1;
}

static int ortp_com_delete(struct sender_command_data *scd)
{
	char *a = para_strdup(inet_ntoa(scd->addr));
	struct ortp_target *ot, *tmp;
	list_for_each_entry_safe(ot, tmp, &targets, node) {
		if (scd->port != ot->port)
			continue;
		if (strcmp(TARGET_ADDR(ot), a))
			continue;
		ortp_delete_target(ot, "com_delete");
	}
	return 1;
}

static void ortp_add_target(int port, struct in_addr *addr)
{
	struct ortp_target *ot = para_malloc(sizeof(struct ortp_target));
	ot->session = NULL;
	ot->port = port;
	ot->addr = *addr;
	PARA_INFO_LOG("adding to target list (%s:%d)\n",
		TARGET_ADDR(ot), ot->port);
	list_add(&ot->node, &targets);
}

static int ortp_com_add(struct sender_command_data *scd)
{
	int port = (scd->port > 0)? scd->port : conf.ortp_default_port_arg;
	ortp_add_target(port, &scd->addr);
	return 1;
}

static char *ortp_info(void)
{
	struct ortp_target *ot;
	char *ret, *tgts = NULL;

	list_for_each_entry(ot, &targets, node) {
		char *tmp = make_message("%s%s:%d ", tgts? tgts : "",
			TARGET_ADDR(ot), ot->port);
		free(tgts);
		tgts = tmp;
	}
	ret = make_message(
		"ortp status: %s\n"
		"ortp default port: udp %d\n"
		"ortp targets: %s\n",
		(self->status == SENDER_ON)? "on" : "off",
		conf.ortp_default_port_arg,
		tgts? tgts : "(none)"
	);
	free(tgts);
	return ret;
}

static int ortp_init_target_list(void)
{
	int i;

	INIT_LIST_HEAD(&targets);
	for (i = 0; i < conf.ortp_target_given; i++) {
		char *arg = para_strdup(conf.ortp_target_arg[i]);
		char *p = strchr(arg, ':');
		int port;
		struct in_addr addr;

		if (!p)
			goto err;
		*p = '\0';
		if (!inet_aton(arg, &addr))
			goto err;
		port = atoi(++p);
		if (port < 0 || port > 65535)
			port = conf.ortp_default_port_arg;
		free(arg);
		ortp_add_target(port, &addr);
		continue;
err:
		PARA_CRIT_LOG("syntax error for ortp_target option #%d\n", i);
		free(arg);
		return -1;
	}
	return 1;
}

static int ortp_set_fds(struct audio_format *af, int *max_fileno, fd_set *rfds, fd_set *wfds)
{
	static int count;
	if (!af)
		return 1;
	int mod = conf.ortp_header_interval_arg / (tv2ms(&af->chunk_tv) + 1);

	if (!(count++ % mod))
		ortp_send_header(af);
	return 1;
}

static char *ortp_help(void)
{
	return make_message(
		"usage: {on|off}\n"
		"usage: {add|delete} IP port\n"
		"example: add 224.0.1.38 1500 (LAN-streaming)\n"
	);
}

int para_ortp_init(struct sender *s)
{
	ortp_init();
	ortp_scheduler_init();
	INIT_LIST_HEAD(&targets);
	s->info = ortp_info;
	s->help = ortp_help;
	s->send = ortp_send;
	s->set_fds = ortp_set_fds;
	s->handle_fds = NULL;
	s->shutdown_clients = ortp_shutdown_targets;
	s->client_cmds[SENDER_ON] = ortp_com_on;
	s->client_cmds[SENDER_OFF] = ortp_com_off;
	s->client_cmds[SENDER_DENY] = NULL;
	s->client_cmds[SENDER_ALLOW] = NULL;
	s->client_cmds[SENDER_ADD] = ortp_com_add;
	s->client_cmds[SENDER_DELETE] = ortp_com_delete;
	self = s;
	s->status = SENDER_OFF;
	if (ortp_init_target_list() < 0)
		return -1;
	if (!conf.ortp_no_autostart_given)
		s->status = SENDER_ON;
	PARA_DEBUG_LOG("%s", "ortp sender init complete\n");
	return 1;
}
